import argparse
import sys

from .scaffolding import read_csv_and_write_output
from .utils import get_currency_choices

__author__ = "Artur Barseghyan"
__copyright__ = "2021 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = (
    "generate_currencies",
    "list_currencies",
)


def generate_currencies():
    """Generate currency modules."""
    parser = argparse.ArgumentParser(description="Generate currency modules")
    parser.add_argument(
        "--in-file",
        dest="in_file",
        default="list_of_circulating_currencies.csv",
        action="store",
        help="Input file",
    )
    parser.add_argument(
        "--out-dir",
        dest="out_dir",
        default="currencies",
        action="store",
        help="Output directory",
    )
    parser.add_argument(
        "--skip-first-line",
        dest="skip_first_line",
        default=False,
        action="store_true",
        help="Skip first line from the CSV?",
    )
    args = parser.parse_args(sys.argv[1:])
    in_file = args.in_file
    out_dir = args.out_dir
    skip_first_line = args.skip_first_line
    read_csv_and_write_output(in_file, out_dir, skip_first_line)


def list_currencies():
    """List currency modules."""
    parser = argparse.ArgumentParser(description="List currency modules")
    parser.add_argument(
        "--sort-by-value",
        dest="sort_by_value",
        default=False,
        action="store_true",
        help="Sort by value?",
    )
    args = parser.parse_args(sys.argv[1:])
    sort_by_key = not args.sort_by_value
    currencies = get_currency_choices(sort_by_key=sort_by_key)
    total = len(currencies)
    print("┌───────────┬──────────────────────────────────────────┐")
    print("│ ISO code  │ Currency                                 │")
    print("├───────────┼──────────────────────────────────────────┤")
    for counter, (iso_code, currency) in enumerate(currencies):
        print(f"│ {iso_code}       │ {currency}{' '*(41-len(currency))}│")
        if counter == total - 1:
            print("└───────────┴──────────────────────────────────────────┘")
        else:
            print("├───────────┼──────────────────────────────────────────┤")



import csv
import os.path

from .helpers import project_dir

__author__ = "Artur Barseghyan"
__copyright__ = "2021 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = (
    "generate_contents",
    "read_csv_and_write_output",
)


def generate_contents(
    iso_code: str,
    number_to_basic: str,
    currency: str,
) -> str:
    """Generate the contents for the given currency."""

    return f'''from ..base import BaseCurrency

__author__ = "Artur Barseghyan"
__copyright__ = "2021 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = ("{iso_code}",)


class {iso_code}(BaseCurrency):
    """{iso_code} - {currency}."""

    uid: str = "{iso_code}"
    rate: int = {number_to_basic}
'''


def read_csv_and_write_output(
    in_file: str = "list_of_circulating_currencies.csv",
    out_dir: str = "currencies",
    skip_first_line: bool = False,
) -> str:
    """Read CSV."""
    if not os.path.isabs(in_file):
        in_file = project_dir(in_file)

    if not os.path.isabs(out_dir):
        out_dir = project_dir(out_dir)

    processed = set([])
    with open(in_file) as csv_file:
        reader = csv.reader(csv_file)
        if skip_first_line:
            next(reader)

        for row in reader:
            currency = row[1].strip()
            iso_code = row[3].strip()
            number_to_basic = row[5].strip()

            if "none" in iso_code:
                continue

            if "none" in number_to_basic:
                number_to_basic = 1

            iso_code = iso_code[:3]

            if iso_code not in processed:
                contents = generate_contents(
                    iso_code, number_to_basic, currency
                )
                with open(f"{out_dir}/{iso_code.lower()}.py", "w") as output:
                    output.write(contents)
                    processed.add(iso_code)

    return out_dir



import warnings
from .base import Registry

warnings.warn(
    "The `Registry` class is moved from `valuta.registry` to `valuta.base`.",
    DeprecationWarning
)

__author__ = "Artur Barseghyan"
__copyright__ = "2021 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = ("Registry",)



from decimal import Decimal
import operator
from typing import Optional, Dict, Union, List, Set, Tuple
from babel.numbers import get_currency_name, get_currency_symbol

from .exceptions import ImproperlyConfigured
from .helpers import classproperty

__author__ = "Artur Barseghyan"
__copyright__ = "2021 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = (
    "BaseCurrency",
    "Registry",
)


class Registry(type):
    REGISTRY: Dict[str, "BaseCurrency"] = {}

    def __new__(mcs, name, bases, attrs):
        new_cls = type.__new__(mcs, name, bases, attrs)
        # Here the name of the class is used as key but it could be any class
        # parameter.
        if getattr(new_cls, "_uid", None):
            mcs.REGISTRY[new_cls._uid] = new_cls
        if new_cls.__name__ != "BaseCurrency":
            new_cls.validate()
        return new_cls

    @property
    def _uid(cls) -> str:
        return getattr(cls, "uid", cls.__name__)

    @classmethod
    def reset(mcs) -> None:
        mcs.REGISTRY = {}

    @classmethod
    def get(mcs, key, default=None) -> Union["BaseCurrency", None]:
        return mcs.REGISTRY.get(key, default)

    @classmethod
    def items(mcs) -> List[Tuple[str, "BaseCurrency"]]:
        return mcs.REGISTRY.items()

    @classmethod
    def values(
        mcs,
        limit_choices_to: Union[Tuple[str, ...], List[str], Set[str]] = None,
        sort_by_key: bool = False,
    ):
        if limit_choices_to is None:
            values = [
                (__key, __value.name)
                for __key, __value in mcs.REGISTRY.items()
            ]
        else:
            values = [
                (__key, __value.name)
                for __key, __value in mcs.REGISTRY.items()
                if __key in limit_choices_to
            ]
        if sort_by_key:
            values.sort(key=operator.itemgetter(0))
        else:
            values.sort(key=operator.itemgetter(1))
        return values

    @classmethod
    def values_with_code(
        mcs,
        limit_choices_to: Union[Tuple[str, ...], List[str], Set[str]] = None,
        sort_by_key: bool = False,
    ):
        if limit_choices_to is None:
            values = [
                (__key, f"{__value.name} ({__key})")
                for __key, __value in mcs.REGISTRY.items()
            ]
        else:
            values = [
                (__key, f"{__value.name} ({__key})")
                for __key, __value in mcs.REGISTRY.items()
                if __key in limit_choices_to
            ]
        if sort_by_key:
            values.sort(key=operator.itemgetter(0))
        else:
            values.sort(key=operator.itemgetter(1))
        return values


class BaseCurrency(metaclass=Registry):
    """Base currency.

    Each currency typically has a main currency unit (the dollar, for
    example, or the euro) and a fractional unit, often defined as 1⁄100 of
    the main unit: 100 cents = 1 dollar, 100 centimes = 1 franc,
    100 pence = 1 pound, although units of 1⁄10 or 1⁄1000 occasionally also
    occur. Some currencies do not have any smaller units at all, such as the
    Icelandic krona.

    https://en.wikipedia.org/wiki/Currency
    """

    uid: Optional[str] = None
    name: str
    rate: int

    @classmethod
    def validate(cls):
        """Constructor."""
        if not hasattr(cls, "rate"):
            raise ImproperlyConfigured(
                "The `rate` property of the currency shall be defined."
            )

    @classmethod
    def convert_to_currency_units(
        cls, value: Union[int, float, Decimal]
    ) -> Union[int, float, Decimal]:
        """Convert to amount in currency units."""
        return value / cls.rate

    @classproperty
    def name(cls):
        """Automatic currency name.

        Example::

            from babel.numbers import get_currency_name
            In: get_currency_name('AMD')
            Out: 'Armenian Dram'

            In: get_currency_name('AMD', locale='hy_AM')
            Out: 'հայկական դրամ'

            In: get_currency_name('AMD', locale='nl_NL')
            Out: 'Armeense dram'

        :return:
        """
        return get_currency_name(cls.uid)

    @classproperty
    def symbol(cls):
        """Automatic currency symbol.

        Example::

            from babel.numbers import get_currency_symbol
            In: get_currency_symbol('AMD')
            Out: 'AMD'

            In: get_currency_symbol('AMD', locale='hy_AM')
            Out: '֏'

            In: get_currency_symbol('AMD', locale='nl_NL')
            Out: 'AMD'

        :return:
        """
        return get_currency_symbol(cls.uid)



from ..base import BaseCurrency

__author__ = "Artur Barseghyan"
__copyright__ = "2021 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = ("LKR",)


class LKR(BaseCurrency):
    """LKR - Sri Lankan rupee."""

    uid: str = "LKR"
    rate: int = 100



import unittest

from ..utils import get_currency_choices, get_currency_choices_with_code
from .data import (
    CURRENCY_CHOICES,
    CURRENCY_CHOICES_SORT_BY_KEY,
    CURRENCY_CHOICES_WITH_CODE,
    CURRENCY_CHOICES_WITH_CODE_SORT_BY_KEY,
)

__author__ = "Artur Barseghyan"
__copyright__ = "2021 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = ("TestUtils",)


class TestUtils(unittest.TestCase):
    """Test utils."""

    def setUp(self):
        """Set up."""

    def test_get_currency_choices(self):
        """Test get_currency_choices."""
        choices = get_currency_choices()
        self.assertListEqual(choices, CURRENCY_CHOICES)

    def test_get_currency_choices_sort_by_key(self):
        """Test get_currency_choices sort_by_key=True."""
        choices = get_currency_choices(sort_by_key=True)
        self.assertListEqual(choices, CURRENCY_CHOICES_SORT_BY_KEY)

    def test_get_currency_choices_with_code(self):
        """Test get_currency_choices_with_code."""
        choices = get_currency_choices_with_code()
        self.assertListEqual(choices, CURRENCY_CHOICES_WITH_CODE)

    def test_get_currency_choices_with_code_sort_by_key(self):
        """Test get_currency_choices_with_code sort_by_key=True."""
        choices = get_currency_choices_with_code(sort_by_key=True)
        self.assertListEqual(choices, CURRENCY_CHOICES_WITH_CODE_SORT_BY_KEY)



import unittest

from ..currencies import TND, EUR, MRU, VND, UGX
from ..exceptions import InvalidCurrency
from ..shortcuts import convert_to_currency_units


__author__ = "Artur Barseghyan"
__copyright__ = "2021 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = ("TestShortcuts",)


class TestShortcuts(unittest.TestCase):
    """Test shortcuts."""

    def test_convert_to_currency_units(self):
        """Test `convert_to_currency_units` with valid codes."""
        self.assertEqual(
            convert_to_currency_units(
                UGX.uid,
                10
            ),
            10
        )

        self.assertEqual(
            convert_to_currency_units(
                MRU.uid,
                5
            ),
            1
        )

        self.assertEqual(
            convert_to_currency_units(
                VND.uid,
                10
            ),
            1
        )

        self.assertEqual(
            convert_to_currency_units(
                EUR.uid,
                100
            ),
            1
        )

        self.assertEqual(
            convert_to_currency_units(
                TND.uid,
                1_000
            ),
            1
        )

    def _test_currency_code_is_none(self, fail_silently: bool = True):
        self.assertEqual(
            convert_to_currency_units(
                None,
                1_000,
                fail_silently
            ),
            None
        )

    def _test_currency_code_is_empty_string(self, fail_silently: bool = True):
        self.assertEqual(
            convert_to_currency_units(
                "",
                1_000,
                fail_silently
            ),
            None
        )

    def _test_currency_code_is_invalid(self, fail_silently: bool = True):
        self.assertEqual(
            convert_to_currency_units(
                "None",
                1_000,
                fail_silently
            ),
            None
        )

    def test_test_convert_to_currency_units_invalid_currency_code(self):
        """Test `convert_to_currency_units` with invalid codes."""

        with self.subTest("currency_code=None"):
            self._test_currency_code_is_none()

        with self.subTest("currency_code=''"):
            self._test_currency_code_is_empty_string()

        with self.subTest("currency_code='None'"):
            self._test_currency_code_is_invalid()

        with self.assertRaises(InvalidCurrency):
            with self.subTest("currency_code=None"):
                self._test_currency_code_is_none(fail_silently=False)

        with self.assertRaises(InvalidCurrency):
            with self.subTest("currency_code=''"):
                self._test_currency_code_is_empty_string(fail_silently=False)

        with self.assertRaises(InvalidCurrency):
            with self.subTest("currency_code='None'"):
                self._test_currency_code_is_invalid(fail_silently=False)



import os

__author__ = "Artur Barseghyan"
__copyright__ = "2021 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = (
    "ClassProperty",
    "classproperty",
    "project_dir",
)


class ClassProperty(property):
    """ClassProperty.

    How to use::

        class Something:
            @classproperty
            def default_value(cls):
                return 10  # Some value

    Can be replaced with the following in Python 3.9::

        class Something:
            @classmethod
            @property
            def default_value(cls):
                return 10  # Some value
    """

    def __get__(self, cls, owner):
        """Get."""
        return classmethod(self.fget).__get__(None, owner)()


classproperty = ClassProperty


def project_dir(base: str) -> str:
    """Absolute path to a file from current directory."""
    return os.path.abspath(
        os.path.join(os.path.dirname(__file__), base).replace("\\", "/")
    )



from decimal import Decimal
from typing import Union

from .base import Registry
from .exceptions import InvalidCurrency

__author__ = "Artur Barseghyan"
__copyright__ = "2021 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = ("convert_to_currency_units",)


def convert_to_currency_units(
    currency_code: str, value: int, fail_silently: bool = True
) -> Union[int, float, Decimal]:
    """Convert value represented in minor currency to major currency units."""
    cls = Registry.get(currency_code)
    if cls:
        return cls.convert_to_currency_units(value)
    elif not fail_silently:
        raise InvalidCurrency(
            f"Invalid or empty currency code: {currency_code}"
        )



from ..base import BaseCurrency

__author__ = "Artur Barseghyan"
__copyright__ = "2021 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = ("BOB",)


class BOB(BaseCurrency):
    """BOB - Bolivian boliviano."""

    uid: str = "BOB"
    rate: int = 100
